use std::{collections::HashMap, fmt};

use crate::PreLispValue;

#[cfg(test)]
mod tests {
    use crate::{parse, LispState, LispValue, Tokenizer};

    use super::add_function;

	fn get_exec_result(source: &str) -> Result<Vec<LispValue>, Box<dyn std::error::Error>> {
		let mut state = LispState::new();

		state.table.insert(String::from("output"), LispValue::List(vec![], false));
		add_function(&mut state.table, "print", |st, args| {
			if let LispValue::List(v, _) = st.table.get_mut("output").unwrap() {
				v.push(LispValue::List(args, false))
			}

			Ok(vec![LispValue::Nil])
		});

		let mut tokenizer = Tokenizer::new(String::from(source));
		let instructions = parse(tokenizer.tokenize()?);

		state.execute(instructions)?;

		let mut out = Vec::new();
		if let LispValue::List(outerlist, _) = &state.table["output"] {
			for list in outerlist {
				if let LispValue::List(innerlist, _) = list {
					for innervalue in innerlist {
						out.push(innervalue.to_owned());
					}
				}
			}
		}

		Ok(out)
	}

	#[test]
	fn test_hello_world() -> Result<(), Box<dyn std::error::Error>> {
		assert_eq!(
			vec![
				LispValue::String(String::from("Hello, World!"))
			],
			get_exec_result("
				(print \"Hello, World!\")
			")?
		);
		Ok(())
	}

	#[test]
	fn test_math() -> Result<(), Box<dyn std::error::Error>> {
		assert_eq!(
			vec![
				LispValue::Float(16.05)
			],
			get_exec_result("
				(print 
					(add 
						(div
							4.5
							(sub 0.5 0.2)
						)
						(mul 21 0.05)
					)
				)
			")?
		);
		Ok(())
	}
}

pub struct LispState {
	table: HashMap<String, LispValue>,
}


#[derive(Debug)]
pub struct RuntimeError {
	message: String,
}

impl fmt::Display for RuntimeError {
	fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
		write!(f, "Runtime Error: {}", self.message)
	}
}

impl std::error::Error for RuntimeError {}

type RustFunctionType = fn(&mut LispState, Vec<LispValue>) -> Result<Vec<LispValue>, RuntimeError>;

#[derive(Debug, Clone, PartialEq)]
pub enum LispValue {
	Nil,
	String(String),
	Integer(i32),
	Float(f32),
	List(Vec<LispValue>, bool),
	LispFunction(String, Vec<String>, Vec<LispValue>),
	RustFunction(String, RustFunctionType),
	Var(String)
}

impl fmt::Display for LispValue {
	fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
		#[allow(unreachable_patterns)]
		match self {
			LispValue::Nil => write!(f, "nil"),
			LispValue::String(str) => write!(f, "\"{}\"", str),
			LispValue::LispFunction(name, _, _) => write!(f, "<function {}>", name),
			LispValue::RustFunction(name, _) => write!(f, "<function {}>", name),
			LispValue::Integer(num) => write!(f, "{}", num),
			LispValue::Float(num) => write!(f, "{}", num),
			LispValue::List(elems, quoted) => {
				let mut strs = Vec::new();

				for e in elems {
					strs.push(e.to_string());
				}

				if *quoted {
					write!(f, "'({})", strs.join(" "))
				} else {
					write!(f, "({})", strs.join(" "))
				}
			},
			val => {
				write!(f, "{}", format!("{:?}", val))
			}
		}
	}
}

fn get_type(v: &LispValue) -> String {
	match v {
		LispValue::Nil => String::from("nil"),
		LispValue::Integer(_) => String::from("int"),
		LispValue::Float(_) => String::from("float"),
		LispValue::String(_) => String::from("string"),
		LispValue::List(_, _) => String::from("list"),
		LispValue::LispFunction(_, _, _) => String::from("function"),
		LispValue::RustFunction(_, _) => String::from("function"),
		LispValue::Var(_) => String::from("symbol")
	}
}

impl From<&str> for LispValue {
	fn from(value: &str) -> Self {
		LispValue::String(value.to_string())
	}
}
impl From<String> for LispValue {
	fn from(value: String) -> Self {
		LispValue::String(value)
	}
}
impl From<i32> for LispValue {
	fn from(value: i32) -> Self {
		LispValue::Integer(value)
	}
}
impl From<f32> for LispValue {
	fn from(value: f32) -> Self {
		LispValue::Float(value)
	}
}
impl From<Vec<LispValue>> for LispValue {
	fn from(value: Vec<LispValue>) -> Self {
		LispValue::List(value, false)
	}
}

fn add_function<T: Into<String> + Clone>(table: &mut HashMap<String, LispValue>, name: T, func: RustFunctionType) {
	table.insert(name.clone().into(), LispValue::RustFunction(name.into(), func));
}

impl LispState {
	pub fn new() -> LispState {
		let mut table = HashMap::new();

		add_function(&mut table, "print", |_, args| {
			let mut strings = Vec::new();
			for val in args {
				strings.push(val.to_string());
			}

			let str = strings.join(" ");
			println!("{}", str);

			Ok(vec![LispValue::Nil])
		});

		add_function(&mut table, "concat", |_, args| {
			let mut strings = Vec::new();
			for val in args {
				strings.push(val.to_string());
			}

			let str = strings.concat();
			Ok(vec![LispValue::String(str)])
		});

		add_function(&mut table, "add", |_, args| {
			if args.len() < 1 {
				return Ok(vec![LispValue::Nil])
			} else if args.len() == 1 {
				return Ok(vec![args[0].clone()]);
			}

			let mut t = "int";
			for v in &args {
				match v {
					LispValue::Float(_) => t = "float",
					LispValue::Integer(_) => (),
					_ => return Err(RuntimeError {
						message: format!("Invalid argument to 'add' (expected float or int, got {}", get_type(v))
					})
				}
			}

			match t {
				"int" => {
					let mut sum = 0;
					for v in &args {
						if let LispValue::Integer(n) = v {
							sum += n;
						} else {
							unreachable!()
						}
					}
					return Ok(vec![LispValue::Integer(sum)]);
				},
				"float" => {
					let mut sum = 0.0 as f32;
					for v in &args {
						if let LispValue::Float(n) = v {
							sum += n;
						} else if let LispValue::Integer(n) = v{
							sum += *n as f32;
						} else {
							unreachable!()
						}
					}
					return Ok(vec![LispValue::Float(sum)]);
				},
				_ => unreachable!()
			}
		});

		add_function(&mut table, "sub", |_, args| {
			if args.len() < 1 {
				return Ok(vec![LispValue::Nil]);
			} else if args.len() == 1 {
				return Ok(vec![args[0].clone()]);
			}

			let mut t = "int";
			for v in &args {
				match v {
					LispValue::Float(_) => t = "float",
					LispValue::Integer(_) => (),
					_ => return Err(RuntimeError {
						message: format!("Invalid argument to 'sub' (expected float or int, got {}", get_type(v))
					})
				}
			}

			match t {
				"int" => {
					if let LispValue::Integer(n) = args[0] {
						let mut sum = n;
						for i in 1..args.len() {
							if let LispValue::Integer(n) = &args[i] {
								sum -= n;
							} else {
								unreachable!()
							}
						}
						return Ok(vec![LispValue::Integer(sum)]);
					} else {
						unreachable!()
					}
				},
				"float" => {
					if let LispValue::Float(n) = args[0] {
						let mut sum = n;
						for i in 1..args.len() {
							if let LispValue::Float(n) = &args[i] {
								sum -= *n;
							} else if let LispValue::Integer(n) = &args[i] {
								sum -= *n as f32;
							} else {
								unreachable!()
							}
						}
						return Ok(vec![LispValue::Float(sum)]);
					} else if let LispValue::Integer(n) = args[0] {
						let mut sum = n as f32;
						for i in 1..args.len() {
							if let LispValue::Float(n) = &args[i] {
								sum -= *n;
							} else if let LispValue::Integer(n) = &args[i] {
								sum -= *n as f32;
							} else {
								unreachable!()
							}
						}
						return Ok(vec![LispValue::Float(sum)]);
					} {
						unreachable!()
					}
				},
				_ => unreachable!()
			}
		});

		add_function(&mut table, "mul", |_, args| {
			if args.len() < 1 {
				return Ok(vec![LispValue::Nil]);
			} else if args.len() == 1 {
				return Ok(vec![args[0].clone()]);
			}

			let mut t = "int";
			for v in &args {
				match v {
					LispValue::Float(_) => t = "float",
					LispValue::Integer(_) => (),
					_ => return Err(RuntimeError {
						message: format!("Invalid argument to 'mul' (expected float or int, got {}", get_type(v))
					})
				}
			}

			match t {
				"int" => {
					if let LispValue::Integer(n) = args[0] {
						let mut sum = n;
						for i in 1..args.len() {
							if let LispValue::Integer(n) = &args[i] {
								sum *= n;
							} else {
								unreachable!()
							}
						}
						return Ok(vec![LispValue::Integer(sum)]);
					} else {
						unreachable!()
					}
				},
				"float" => {
					if let LispValue::Float(n) = args[0] {
						let mut product = n;
						for i in 1..args.len() {
							if let LispValue::Float(n) = &args[i] {
								product *= *n;
							} else if let LispValue::Integer(n) = &args[i] {
								product *= *n as f32;
							} else {
								unreachable!()
							}
						}
						return Ok(vec![LispValue::Float(product)]);
					} else if let LispValue::Integer(n) = args[0] {
						let mut product = n as f32;
						for i in 1..args.len() {
							if let LispValue::Float(n) = &args[i] {
								product *= *n;
							} else if let LispValue::Integer(n) = &args[i] {
								product *= *n as f32;
							} else {
								unreachable!()
							}
						}
						return Ok(vec![LispValue::Float(product)]);
					} {
						unreachable!()
					}
				},
				_ => unreachable!()
			}
		});

		add_function(&mut table, "div", |_, args| {
			if args.len() < 1 {
				return Ok(vec![LispValue::Nil]);
			} else if args.len() == 1 {
				return Ok(vec![args[0].clone()]);
			}

			let mut t = "int";
			for v in &args {
				match v {
					LispValue::Float(_) => t = "float",
					LispValue::Integer(_) => (),
					_ => return Err(RuntimeError {
						message: format!("Invalid argument to 'div' (expected float or int, got {}", get_type(v))
					})
				}
			}

			match t {
				"int" => {
					if let LispValue::Integer(n) = args[0] {
						let mut sum = n;
						for i in 1..args.len() {
							if let LispValue::Integer(n) = &args[i] {
								sum /= n;
							} else {
								unreachable!()
							}
						}
						return Ok(vec![LispValue::Integer(sum)]);
					} else {
						unreachable!()
					}
				},
				"float" => {
					if let LispValue::Float(n) = args[0] {
						let mut product = n;
						for i in 1..args.len() {
							if let LispValue::Float(n) = &args[i] {
								product /= *n;
							} else if let LispValue::Integer(n) = &args[i] {
								product /= *n as f32;
							} else {
								unreachable!()
							}
						}
						return Ok(vec![LispValue::Float(product)]);
					} else if let LispValue::Integer(n) = args[0] {
						let mut product = n as f32;
						for i in 1..args.len() {
							if let LispValue::Float(n) = &args[i] {
								product /= *n;
							} else if let LispValue::Integer(n) = &args[i] {
								product /= *n as f32;
							} else {
								unreachable!()
							}
						}
						return Ok(vec![LispValue::Float(product)]);
					} {
						unreachable!()
					}
				},
				_ => unreachable!()
			}
		});

		LispState {
			table
		}
	}

	fn handle_prevalues(&self, input: Vec<PreLispValue>) -> Vec<LispValue> {
		let mut out = Vec::new();

		for v in input {
			out.push(match v {
				PreLispValue::Nil => LispValue::Nil,
				PreLispValue::String(str) => LispValue::String(str),
				PreLispValue::Integer(num) => LispValue::Integer(num),
				PreLispValue::Float(num) => LispValue::Float(num),
				PreLispValue::List(list, quoted) => {
					LispValue::List(Self::handle_prevalues(self, list), quoted)
				},
				PreLispValue::LispFunction(name, arg_names, instructions) => {
					LispValue::LispFunction(name, arg_names, Self::handle_prevalues(self, instructions))
				},
				PreLispValue::Var(name) => {
					LispValue::Var(name)
				}
			})
		}

		out
	}

	fn eval_list(&mut self, env: &mut HashMap<String, LispValue>, list: &Vec<LispValue>, quoted: bool) -> Result<(Vec<LispValue>, bool), RuntimeError> {
		if list.len() < 1 {
			return Ok((vec![LispValue::List(Vec::new(), quoted)], false));
		}

		let list = &mut list.clone();

		for i in 0..list.len() {
			if let LispValue::Var(name) = &list[i] {
				list[i] = env.get(name).unwrap_or_else(|| self.table.get(name).unwrap_or(&LispValue::Nil)).to_owned();
			}
			if let LispValue::List(elements, quoted) = &list[i] {
				if let Ok((elems, astuple)) = self.eval_list(env, elements, *quoted) {
					if astuple {
						list[i] = elems.get(0).unwrap_or(&LispValue::Nil).to_owned();
						for j in 1..elems.len() {
							list.insert(i, elems[j].to_owned());
						}
					} else {
						list[i] = LispValue::List(elems, *quoted);
					}
				}
			}
		}

		if quoted {
			return Ok((list.to_vec(), false));
		}

		if let LispValue::RustFunction(_name, f) = &list[0] {
			Ok((f(self, list[1..].to_vec())?, true))
		} else if let LispValue::LispFunction(_name, _, _) = &list[0] {
			Ok((self.call_lisp_func(env, &list[0], list[1..].to_vec())?, true))
		} else {
			Ok((list.to_vec(), false))
		}
	}

	fn call_lisp_func(&mut self, env: &mut HashMap<String, LispValue>, f: &LispValue, args: Vec<LispValue>) -> Result<Vec<LispValue>, RuntimeError> {
		let (arg_names, instructions) = match f {
			LispValue::LispFunction(_, arg_names, i) => {
				(arg_names, i)
			},
			_ => return Err(RuntimeError {
				message: format!("Invalid argument to call_lisp_func (expected function, got {})", get_type(f))
			})
		};

		for i in 0..arg_names.len() {
			env.insert(arg_names[i].to_owned(), args.get(i).unwrap_or(&LispValue::Nil).to_owned());
		}

		let mut out = Vec::new();
		for l in instructions {
			if let LispValue::List(elements, quoted) = l {
				let (values, astuple) = self.eval_list(env, &elements, *quoted)?;
				if values.len() > 0 && values[0] != LispValue::Nil {
					if astuple {
						for v in values {
							out.push(v);
						}
					} else {
						out.push(values.into());
					}
				}
			}
		}
		Ok(out)
	}

	pub fn execute(&mut self, instructions: Vec<PreLispValue>) -> Result<Vec<LispValue>, RuntimeError> {
		let instructions = Self::handle_prevalues(self, instructions);
		
		let mut outs = Vec::new();

		let mut env = HashMap::new();

		for val in instructions {
			if let LispValue::List(elements, quoted) = val {
				let (vals, astuple) = self.eval_list(&mut env, &elements, quoted)?;
				if astuple {
					for val in vals {
						outs.push(val);
					}
				} else {
					outs.push(vals.into())
				}
			} else {
				outs.push(val);
			}
		}

		Ok(outs)
	}
}
